import logging

import sys
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import boto3
from typer import Argument, Option, Typer

from dsnap import utils
from dsnap.snapshot import Snapshot
from dsnap.utils import ask_to_create_snapshot, full_prompt

if TYPE_CHECKING:
    from mypy_boto3_ec2 import service_resource as r

app = Typer()

sess: boto3.session.Session = boto3.session.Session()

# This gets set via @app.callback before any command runs.
ec2: 'r.EC2ServiceResource' = None  # type: ignore[assignment]


@app.callback()
def session(region: str = Option(default='us-east-1'), profile: str = Option(default=None)):
    """This is function set's up various global settings.

    It is called by Typer before any of the other commands run due to the @app.callback decorator.
    """
    global sess, ec2
    sess = boto3.session.Session(region_name=region, profile_name=profile)
    ec2 = sess.resource('ec2')


@app.command()
def init():
    """Initializes the current working directory with a templated Vagrantfile for mounting downloaded images"""
    template = Path(__file__).parent.joinpath(Path('templates/Vagrantfile'))
    out = Path('Vagrantfile')
    if out.exists():
        logging.error("This directory already contains a Vagrantfile.")
        sys.exit(101)
    else:
        out.write_text(template.read_text())


@app.command("list")
def list_snapshots():
    print("           Id          |   Owneer ID   | Description")
    ec2: 'r.EC2ServiceResource' = sess.resource('ec2')
    for snap in ec2.snapshots.filter(OwnerIds=['self']).all():
        print(f"{snap.id}   {snap.owner_id}   {snap.description}")


@app.command()
def get(id: str = Argument(None), output: Optional[Path] = None):
    if id.startswith('snap-'):
        snap_id = id
    elif id.startswith('i-'):
        vol = utils.volume_prompt(ec2.Instance(id).volumes)
        snap_id = (utils.snapshot_prompt(vol.snapshots) or ask_to_create_snapshot(vol)).snapshot_id
    elif not id:
        snap_id = full_prompt(sess)
        if not snap_id:
            print("Exiting...")
    else:
        print("Unknown argument type, first argument should be an Instance Id or Snapshot Id")
        sys.exit(1)

    try:
        logging.info(f"Selected snapshot with id {snap_id}")
        snap = Snapshot(snap_id, boto3_session=sess)
        path = output and output.absolute().as_posix()
        snap.download(path or f"{id}.img")
    except UserWarning as e:
        print(*e.args)
        sys.exit(2)
    except Exception as e:
        resp = getattr(e, 'response', None)
        if resp and resp['Error']['Message']:
            print(resp['Error']['Message'])
            sys.exit(1)
        else:
            raise e



import atexit
import hashlib
import json
import logging
import signal
from base64 import b64encode

import boto3
import sys
from typing import List, Iterable, Dict, cast, Optional
import jmespath

from typing import TYPE_CHECKING

from boto3.resources.base import ServiceResource
from boto3.resources.collection import ResourceCollection

if TYPE_CHECKING:
    from mypy_boto3_ec2 import service_resource as r


def get_tag(tags: Iterable[Dict[str, str]], key: str) -> str:
    """Takes a list of tags and a key name, returns the the value for the tag with the given key name."""
    if not tags:
        return ''
    name_tag = filter(lambda t: t['Key'] == key, tags)
    return next(map(lambda t: t['Value'], name_tag), '')


def get_name_tag(tags: List[dict]) -> str:
    """Takes a list of tags and returns the value of the Name tag."""
    return get_tag(tags, 'Name')


def full_prompt(sess: boto3.Session):
    """Prompts the user for all information.

    This is run when dsnap get is run without any options. First we prompt for the EC2
    instance to run against, prompt again if there's if the instance has multiple
    volumes, prompt again for snapshot if volume has multiple snapshots.
    """
    ec2: 'r.EC2ServiceResource' = sess.resource('ec2')
    inst = instance_prompt(ec2.instances.filter(
        Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
    ))
    try:
        vol = volume_prompt(inst.volumes)
        snap = snapshot_prompt(vol.snapshots) or ask_to_create_snapshot(vol)
        return snap and snap.snapshot_id
    except UserWarning as e:
        print(*e.args)
        sys.exit(1)


def item_prompt(collection: ResourceCollection, jmespath_msg: str = None) -> ServiceResource:
    """Prompt's the user for an item to select from the items passed. Item is expected to support the Item protocol."""
    items = list(collection.all())
    if len(items) == 0:
        raise UserWarning(f'No items found when calling {collection._py_operation_name}')

    msg = ''
    for i, item in enumerate(items):
        if jmespath_msg:
            # dump and load json to convert datetime and similar to something readable
            data = json.loads(json.dumps(item.meta.data, default=str))
            msg = ', '.join(jmespath.search(jmespath_msg, data=data))
        name = get_name_tag(item.tags)
        print(f"{i}) {item.id} {name and f'Name: {name}, '}({msg})")

    answer = int(input(f'Select {list(items)[0].meta.resource_model.name}: '))

    try:
        return list(items)[answer]
    except IndexError:
        print(f"Invalid selection, valid inputs are 0 through {len(items) - 1}", file=sys.stderr)
        return item_prompt(collection)


# Called if no snapshot_id is specified when running get
def instance_prompt(instances: ResourceCollection) -> 'r.Instance':
    """Prompts the user to select an EC2 Instance of passed in instances"""
    return cast('r.Instance', item_prompt(instances, jmespath_msg='[PrivateDnsName, VpcId, SubnetId]'))


def snapshot_prompt(snapshots: ResourceCollection) -> Optional['r.Snapshot']:
    """Prompts the user to select a EC2 Snapshot of passed in snapshots"""
    snaps = list(snapshots.all())
    if len(snaps) == 0:
        return None
    elif len(snaps) == 1:
        snap = snaps[0]
    else:
        snap = cast('r.Snapshot', item_prompt(snapshots, jmespath_msg='[StartTime, OwnerId, Description]'))
    return snap


def volume_prompt(volumes: ResourceCollection) -> 'r.Volume':
    """Prompts the user to select a Volume of passed in volumes"""
    vols: List['r.Volume'] = list(volumes.all())
    return vols[0] if len(vols) == 1 else cast('r.Volume', item_prompt(volumes, jmespath_msg='Attachments[*].Device'))


def ask_to_run(msg, func):
    resp = input(f'{msg} [y/N]: ')
    if resp.lower() == 'y':
        return func()
    return None


def ask_to_create_snapshot(vol: 'r.Volume') -> 'r.Snapshot':
    """Asks the user if we should to create a temporary snapshot.

    If the answer is Y we start snapshot creation, wait for it to finish and register a function to
    delete the snapshot on exit.
    """
    return ask_to_run("No snapshots found, create one?", lambda: create_tmp_snap(vol))


def cleanup_snap(snap: 'r.Snapshot'):
    def func():
        print(f'Cleaning up snapshot: {snap.id}')
        snap.delete()

    return func


def create_tmp_snap(vol: 'r.Volume') -> 'r.Snapshot':
    """Creates a temporary snapshot that will get deleted when the process exits."""
    instances = ', '.join([f"{a['InstanceId']} {a['Device']}" for a in vol.attachments])
    desc = f'Instance(s): {instances}, Volume: {vol.id}'
    print(f'Creating snapshot for {desc}')
    snap = vol.create_snapshot(
        Description=f'dsnap ({desc})',
        TagSpecifications=[{
            'ResourceType': 'snapshot',
            'Tags': [{'Key': 'dsnap', 'Value': 'true'}]
        }]
    )
    atexit.register(cleanup_snap(snap))
    signal.signal(signal.SIGTERM, lambda sigs, type: sys.exit())
    print("Waiting for snapshot to complete.")
    snap.wait_until_completed()
    logging.info("Snapshot creation finished")
    return snap


def sha256_check(data: bytes, digest: str) -> bool:
    """Runs sha256 on data and compares it to digest, returns true if these values match.

    digest is expected to be a base64 encoded result of the binary digest.
    """
    m = hashlib.sha256()
    m.update(data)
    chksum = b64encode(m.digest()).decode()
    result = chksum == digest
    if not result:
        logging.error(f'Expected checksum {digest} but got {chksum}')
    return result



import logging
import os
import sys
from pathlib import Path
from queue import Queue, Empty
from threading import Thread
from typing import TYPE_CHECKING, List, NamedTuple

import botocore.config
from botocore.response import StreamingBody

from dsnap.utils import sha256_check

if TYPE_CHECKING:
    from mypy_boto3_ebs.client import EBSClient
    from mypy_boto3_ebs.type_defs import BlockTypeDef

import boto3.resources

MEGABYTE: int = 1024 * 1024
GIBIBYTE: int = 1024 * MEGABYTE

FETCH_THREADS = 30


class Block(NamedTuple):
    BlockData: StreamingBody
    Offset: int
    Checksum: str


class Snapshot:
    def __init__(
            self,
            snapshot_id: str,
            boto3_session: boto3.session.Session = boto3.session.Session(region_name='us-east-1'),
            botocore_conf: botocore.config.Config = botocore.config.Config()
    ) -> None:
        self.snapshot_id = snapshot_id
        self.output_file = ''

        self.queue: Queue = Queue()

        # Make sure the number of connections matches the number of threads we run when fetching the EBS snapshot
        ebs_config = botocore.config.Config(max_pool_connections=FETCH_THREADS).merge(botocore_conf)
        self.ebs: EBSClient = boto3_session.client('ebs', config=ebs_config)

        self.volume_size_b = 0
        self.total_blocks = 0
        self.blocks_written = 0
        self.block_size_b = 0

    def get_blocks(self) -> List['BlockTypeDef']:
        """Retrieves the list of blocks for self.snapshot_id.

        This is called by self.download and doesn't need to be called explicitly before hand.
        """
        resp = self.ebs.list_snapshot_blocks(SnapshotId=self.snapshot_id)

        self.block_size_b = resp['BlockSize']
        self.volume_size_b = resp['VolumeSize'] * GIBIBYTE
        logging.info(f"Volume size is {self.volume_size_b}")

        blocks = resp['Blocks']

        while resp.get('NextToken'):
            resp = self.ebs.list_snapshot_blocks(SnapshotId=self.snapshot_id, NextToken=resp.get('NextToken'))
            blocks.extend(resp['Blocks'])

        self.total_blocks = len(blocks)
        logging.info(f"Number of blocks in image: {self.total_blocks}")

        return blocks

    def download(self, output_file: str, force: bool = False) -> None:
        """Downloads self.snapshot_id to the given output_file.

        If force is true output_file will be overwritten.
        """

        assert output_file
        self.get_blocks()

        if Path(output_file).exists() and not force:
            raise UserWarning(f"The output file '{output_file}' already exists.")

        self.output_file = os.path.abspath(output_file)
        self.truncate()

        threads = list()
        for i in range(FETCH_THREADS):
            t = Thread(target=self._write_blocks_worker)
            threads.append(t)
            t.start()

        blocks = self.get_blocks()
        for block in blocks:
            logging.debug(f"Putting block index {block['BlockIndex']} on the queue")
            self.queue.put(block)

        self.queue.join()

        for t in threads:
            t.join()

        print(f"Output Path: {output_file}")

    def truncate(self) -> None:
        """Truncates self.output_file to size self.volume_size_b."""
        with open(self.output_file, 'wb') as f:
            print(f"Truncating file to {self.volume_size_b}", file=sys.stderr)
            f.truncate(self.volume_size_b)
            f.flush()

    def _write_blocks_worker(self) -> None:
        while self.total_blocks != self.blocks_written:
            try:
                block: 'BlockTypeDef' = self.queue.get(timeout=0.2)
                self._write_block(self._fetch_block(block))
                self.blocks_written += 1
                print(f"Saved block {self.blocks_written} of {self.total_blocks}", end='\r', file=sys.stderr)
                self.queue.task_done()
            except Exception as e:
                if isinstance(e, Empty):
                    continue
                else:
                    logging.exception(f"[ERROR] {e.args}")
                    raise e

    def _fetch_block(self, block: 'BlockTypeDef') -> Block:
        logging.debug(f"Getting block index {block['BlockIndex']}")
        resp = self.ebs.get_snapshot_block(
            SnapshotId=self.snapshot_id,
            BlockIndex=block['BlockIndex'],
            BlockToken=block['BlockToken'],
        )

        # BlockIndex is equal to 512 KiB and seek uses bytes.
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ebs.html#EBS.Client.put_snapshot_block
        return Block(
            Offset=block['BlockIndex'] * self.block_size_b,
            BlockData=resp['BlockData'],
            Checksum=resp['Checksum']
        )

    def _write_block(self, block: Block) -> int:
        """Takes a WriteBlock object to write to disk and yields the number of MiB's for each write."""
        logging.debug(f"Writing block at offset {block.Offset}")
        data = block.BlockData.read()

        if not sha256_check(data, block.Checksum):
            raise UserWarning(f"Got block with incorrect checksum at block offset {block.Offset}")

        with os.fdopen(os.open(self.output_file, os.O_RDWR | os.O_CREAT), 'rb+') as f:
            f.seek(block.Offset)
            bytes_written = f.write(data)
            f.flush()
            return bytes_written
